//
// pvpng.c - Routines for PNG support in Purpleview.
// Uses PNGDIB (which uses libpng)
//

#include "pvconf.h"

#if PV_PNG_SUPPORT

#include <windows.h>
#include <tchar.h>

#include "purpleview.h"
#include <pngdib.h>

#include <png.h>
#include <strsafe.h>

int pv_png_save(HWND hwnd, struct iinfo_struct *iinfo, struct einfo_struct *einfo)
{
	PNGDIB *pngdib;
	int ret;
	TCHAR fn[MAX_PATH];

	if(iinfo->img_fmt==FMT_PNG)
		StringCbCopy(fn,sizeof(fn),iinfo->img_filename);
	else
		StringCbCopy(fn,sizeof(fn),_T(""));

	if(!pv_save_as_dialog(hwnd,_T("Save As PNG"),fn,
		_T("PNG image (*.png)\0*.png\0\0"),_T("png")))
	{
		return 0;
	}

	pngdib=pngdib_d2p_init();
	if(!pngdib) return 0;

	pngdib_d2p_set_png_filename(pngdib,fn);

	pngdib_d2p_set_dib(pngdib,iinfo->lpdib,0,NULL,0);
	pngdib_d2p_set_software_id(pngdib,einfo->software_name);

	if(einfo->use_gamma)
		pngdib_d2p_set_gamma_label(pngdib,1,1.0/einfo->monitor_gamma);

	ret=pngdib_d2p_run(pngdib);

	if(ret) {
		pv_mesg(hwnd,MSG_E,_T("Can't save file: %s"),pngdib_get_error_msg(pngdib));
		pngdib_done(pngdib);
		return 0;
	}

	pngdib_done(pngdib);
	return 1;
}

static void* PNGDIB_DECL heapalloc_callback(void *userdata, int memblksize)
{
	return HeapAlloc((HANDLE)userdata,HEAP_ZERO_MEMORY,memblksize);
}

static void PNGDIB_DECL heapfree_callback(void *userdata, void *memblk)
{
	HeapFree((HANDLE)userdata,0,memblk);
}


int pv_png_open(HWND hwnd, struct iinfo_struct *iinfo, struct einfo_struct *einfo)
{
	PNGDIB *pngdib;
	int ret;
	unsigned char bg_r, bg_g, bg_b;
	int res_x,res_y,res_units;
	int bps;

	pngdib=pngdib_p2d_init();
	if(!pngdib) return 0;

	pngdib_set_userdata(pngdib,(void*)einfo->heap);
	pngdib_setcallback_malloc(pngdib,heapalloc_callback,heapfree_callback,NULL);

	pngdib_p2d_set_png_filename(pngdib,iinfo->img_filename);

	pngdib_p2d_set_use_file_bg(pngdib,einfo->use_image_bg?1:0);

	if(einfo->use_custom_bg) {
		pngdib_p2d_set_custom_bg(pngdib,GetRValue(einfo->custom_bg),
			GetGValue(einfo->custom_bg),GetBValue(einfo->custom_bg));
	}
	if(einfo->use_gamma) {
		pngdib_p2d_set_gamma_correction(pngdib,1,einfo->monitor_gamma);
	}

	ret=pngdib_p2d_run(pngdib);

	if(ret) {   /* returns 0 on success */
		pv_mesg(hwnd,MSG_E,_T("Can't load PNG image: %s"),pngdib_get_error_msg(pngdib));
		pngdib_done(pngdib);
		return 0;
	}

	pngdib_p2d_get_dib(pngdib,&iinfo->lpdib,&iinfo->lpdib_size);

	iinfo->colortype     = pngdib_p2d_get_colortype(pngdib);
	iinfo->bitsperpixel  = pngdib_p2d_get_bitsperpixel(pngdib);
	iinfo->interlaced  = pngdib_p2d_get_interlace(pngdib);
	pngdib_p2d_get_palette(pngdib,NULL,NULL,&iinfo->palettecolors);
	iinfo->compression   = COMPR_DEFLATE;

	if(pngdib_p2d_get_bgcolor(pngdib,&bg_r,&bg_g,&bg_b)) {
		iinfo->bgcolor = RGB(bg_r,bg_g,bg_b);
		iinfo->has_bgcolor=1;
		
	}
	else {
		iinfo->has_bgcolor=0;
	}

	iinfo->res_flag=0;
	if(pngdib_p2d_get_density(pngdib,&res_x,&res_y,&res_units)) {
		if(res_units==1)
			iinfo->res_flag=1;
		else
			iinfo->res_flag=2;
		iinfo->res_x = res_x;
		iinfo->res_y = res_y;
	}

	bps = pngdib_p2d_get_bitspersample(pngdib);

	switch(iinfo->colortype) {
	case 0: // gray
		iinfo->bits_gray= bps;
		break;
	case 2: // rgb
		iinfo->bits_red= iinfo->bits_green= iinfo->bits_blue= bps;
		break;
	// case 3: palette n/a
	case 4: // gray+alpha
		iinfo->bits_gray= iinfo->bits_alpha = bps;
		break;
	case 6: // rgba
		iinfo->bits_red= iinfo->bits_green= 
			iinfo->bits_blue= iinfo->bits_alpha= bps;
		break;
	}

	iinfo->img_fmt=FMT_PNG;

	pngdib_done(pngdib);
	return 1;
}


int pv_png_get_copyright(TCHAR *info, int infolen)
{
	TCHAR buf[250];

	StringCchCopy(info,infolen,_T(""));
	StringCbPrintf(buf,sizeof(buf),_T("Uses pngdib version %s\r\n"),pngdib_get_version_string());
	StringCchCat(info,infolen,buf);
#ifdef UNICODE
	StringCbPrintf(buf,sizeof(buf),_T("Uses libpng version %S\r\n"),png_get_libpng_ver(NULL));
#else
	StringCbPrintf(buf,sizeof(buf),"Uses libpng version %s\r\n",png_get_libpng_ver(NULL));
#endif
	StringCchCat(info,infolen,buf);
	StringCbPrintf(buf,sizeof(buf),_T("Uses zlib version %s\r\n\r\n"),_T(ZLIB_VERSION));
	StringCchCat(info,infolen,buf);

	// fixme
	// Should use png_get_copyright(), but would need to handle
	// carriage-control issues.

	return 1;
}

#endif //PV_PNG_SUPPORT
